/*
 * Zynaptic Reaction - An asynchronous programming framework for Java.
 * 
 * Copyright (c) 2009-2012, Zynaptic Limited. All rights reserved. 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 only, as published by
 * the Free Software Foundation. Zynaptic Limited designates this particular
 * file as subject to the "Classpath" exception as provided in the LICENSE
 * file that accompanied this code.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 * 
 * You should have received a copy of the GNU General Public License version 2
 * along with this work; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Please visit www.zynaptic.com or contact reaction@zynaptic.com if you need
 * additional information or have any questions.
 */

package com.zynaptic.reaction.osgi;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.service.log.LogService;

import com.zynaptic.reaction.Reactor;
import com.zynaptic.reaction.util.ReactorLogTarget;

/**
 * Provides support for logging to an optional OSGi logging service.
 * Availability of a suitable OSGi logging bundle is checked on startup and if
 * present this will be used for the entire lifecycle of the Reaction bundle. A
 * fallback logging target may be specified on startup which will be used if a
 * suitable OSGi service is not available.
 * 
 * @author Chris Holgate
 */
public final class ReactorLogOsgiOptional implements ReactorLogTarget {

  // Backup logging target.
  private final ReactorLogTarget backupLogTarget;

  // Local handle on the OSGI logging service tracker.
  private final ServiceTracker<LogService, LogService> logServiceTracker;

  // Reference back to reaction service.
  private final ServiceReference<Reactor> reactorServiceReference;

  /**
   * Implements default constructor for the optional OSGi logging service.
   * Attaches to the OSGi log service if available and falls back to the backup
   * logging target if not.
   * 
   * @param bundleContext
   *          This is the OSGi bundle context associated with the Reaction
   *          bundle.
   * @param backupLogTarget
   *          This is the backup logging target which will be used if the
   *          optional OSGi log service is not available.
   */
  public ReactorLogOsgiOptional(BundleContext bundleContext, ReactorLogTarget backupLogTarget) {
    this.backupLogTarget = backupLogTarget;
    this.logServiceTracker = logServiceTrackerInit(bundleContext);
    this.reactorServiceReference = bundleContext.getServiceReference(Reactor.class);
  }

  /*
   * Initialisation function for OSGi log service tracker.
   */
  private final synchronized ServiceTracker<LogService, LogService> logServiceTrackerInit(
      BundleContext bundleContext) {
    ServiceTracker<LogService, LogService> tracker = null;
    try {
      tracker = new ServiceTracker<LogService, LogService>(bundleContext, LogService.class, null);
      tracker.open();
      backupLogTarget.log(ReactorLogTarget.LOG_INFO,
          "OSGi logging service installed - backup service will not be used.");
    } catch (NoClassDefFoundError exc) {
      if (backupLogTarget != null) {
        backupLogTarget.log(ReactorLogTarget.LOG_WARNING,
            "OSGi logging service not available - using backup log service.", exc);
      }
    }
    return tracker;
  }

  /*
   * Implements ReactorLogTarget.log(...)
   */
  public final synchronized void log(int level, String message) {
    boolean logged = false;
    if (logServiceTracker != null) {
      LogService logger = logServiceTracker.getService();
      if (logger != null) {
        logger.log(reactorServiceReference, level, message);
        logged = true;
      }
    }

    // Fallback to backup log service.
    if (!logged && (backupLogTarget != null)) {
      backupLogTarget.log(level, message);
    }
  }

  /*
   * Implements ReactorLogTarget.log(...)
   */
  public final synchronized void log(int level, String message, Throwable exception) {
    boolean logged = false;
    if (logServiceTracker != null) {
      LogService logger = logServiceTracker.getService();
      if (logger != null) {
        logger.log(reactorServiceReference, level, message, exception);
        logged = true;
      }
    }

    // Fallback to backup log service.
    if (!logged && (backupLogTarget != null)) {
      backupLogTarget.log(level, message, exception);
    }
  }
}