/*
 * Zynaptic Reaction - An asynchronous programming framework for Java.
 * 
 * Copyright (c) 2009-2012, Zynaptic Limited. All rights reserved. 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 only, as published by
 * the Free Software Foundation. Zynaptic Limited designates this particular
 * file as subject to the "Classpath" exception as provided in the LICENSE
 * file that accompanied this code.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 * 
 * You should have received a copy of the GNU General Public License version 2
 * along with this work; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Please visit www.zynaptic.com or contact reaction@zynaptic.com if you need
 * additional information or have any questions.
 */

package com.zynaptic.reaction.util;

/**
 * Defines the API used by the Reaction package to submit log messages to a
 * suitable log service. For example, standard POJO environments can use this
 * interface to wrap <code> java.util.logging </code> and OSGi environments can
 * use it to wrap <code> org.osgi.service.log </code>. Note that the log levels
 * are intentionally chosen to match those used by the OSGi logging service.
 * 
 * @author Chris Holgate
 */
public interface ReactorLogTarget {

  /**
   * Specifies the error message log level (severity 1). A log message at this
   * severity level indicates a failure which may render the reactor service
   * inoperable or unstable.
   */
  public static final int LOG_ERROR = 1;

  /**
   * Specifies the warning message log level (severity 2). A log message at this
   * severity level indicates a condition which may affect the stability or
   * performance of the reactor service.
   */
  public static final int LOG_WARNING = 2;

  /**
   * Specifies the informational message log level (severity 3). Log messages at
   * this severity are used to submit information about normal runtime events to
   * the log.
   */
  public static final int LOG_INFO = 3;

  /**
   * Specifies the debug message log level (severity 4). Log messages at this
   * severity are only used during debug and should be filtered out during
   * normal operation.
   */
  public static final int LOG_DEBUG = 4;

  /**
   * Logs a message string at the specified severity level.
   * 
   * @param level
   *          This is the severity level at which the message should be logged.
   * @param msg
   *          This is the message string which should be included in the log
   *          file.
   */
  public void log(int level, String msg);

  /**
   * Logs message string and accompanying exception information at the specified
   * severity level.
   * 
   * @param level
   *          This is the severity level at which the message should be logged.
   * @param msg
   *          This is the message string which should be included in the log
   *          file.
   * @param thrown
   *          This is the exception which should be logged with the accompanying
   *          message.
   */
  public void log(int level, String msg, Throwable thrown);

}